/* Copyright 2013 Daniel Hanrahan. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE DANIEL HANRAHAN "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL DANIEL HANRAHAN BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef BOF_LIL_ENDIAN_READER_H
#define BOF_LIL_ENDIAN_READER_H

#include <string>
#include <vector>
#include <algorithm>
#include <boost/smart_ptr/shared_ptr.hpp>
#include <SDL_endian.h>
#ifdef __MINGW32__
#undef main // Prevents SDL from overriding main()
#endif

/// Interface for reading data with little endian encoding.
class BofReader {
	// Disable copy and assignment.
	BofReader (const BofReader& other);
	const BofReader& operator= (const BofReader& other);

	/// \ingroup Binary input:
	/// Any subclass MUST implement this function!
	virtual void read (void *dest, size_t n) = 0;

public:
	typedef boost::shared_ptr<BofReader> ptr;

	BofReader () {}
	virtual ~BofReader() {}


	/// How big is the buffer?
	virtual size_t size() const = 0;

	/// Are we at the end of the buffer?
//	virtual bool eof() const = 0;
	virtual bool eof() const {
		return( position() == size() );
	}


	/// \name Positioning:
	/// \{

	/// How far are we from the start of the buffer?
	virtual size_t position() const = 0;

	/// Set the position in the buffer.
	virtual void seek (size_t position) = 0;

	/// Skip a number of bytes from the current position.
	virtual void skip (size_t num_bytes) = 0;

	/// \}


	/// \name Binary input:
	/// \{

	/// Read a byte (8-bit unsigned).
	unsigned char read_byte() {
		unsigned char c;
		read( &c, 1 );
		return c;
	}

	/// Read a character.
	char read_char() {
		char c;
		read( &c, 1 );
		return c;
	}

	/// Read a 16-bit unsigned int.
	unsigned short read_word() {
		Uint16 val;
		read( &val, 2 );
		return SDL_SwapLE16( val );
	}

	/// Read a 16-bit signed int.
	short read_short() {
		union
		{
			Uint16 ui;
			Sint16 si;
		} swapper;
		read( &swapper.ui, 2 );
		swapper.ui = SDL_SwapLE16( swapper.ui );
		return swapper.si;
	}

	/// Read a 32-bit unsigned int.
	unsigned int read_dword() {
		Uint32 val;
		read( &val, 4 );
		return SDL_SwapLE32( val );
	}

	/// Read a 32-bit signed int.
	int read_int() {
		union
		{
			Uint32 ui;
			Sint32 si;
		} swapper;
		read( &swapper.ui, 4 );
		swapper.ui = SDL_SwapLE32( swapper.ui );
		return swapper.si;
	}


	/// Read a boolean, from 8-bit unsigned.
	bool read_bool8() {
		const unsigned val = read_byte();
		assert( val == 0 || val == 1 );
		return (bool)val;
	}

	/// Read a boolean, from 16-bit unsigned.
	bool read_bool16() {
		const unsigned val = read_word();
		assert( val == 0 || val == 1 );
		return (bool)val;
	}

	/// Read a boolean, from 32-bit unsigned.
	bool read_bool32() {
		const unsigned val = read_dword();
		assert( val == 0 || val == 1 );
		return (bool)val;
	}


	/// Read a bit-array, from 8-bit unsigned.
	std::vector<bool> read_bits8() {
		std::vector<bool> bits;
		bits.reserve(8);
		const unsigned val = read_byte();
		for (size_t i = 0; i < 8; ++i) {
			bits.push_back( val & (1u<<i) );
		}
		return bits;
	}

	/// Read a bit-array, from 16-bit unsigned.
	std::vector<bool> read_bits16() {
		std::vector<bool> bits;
		bits.reserve(16);
		const unsigned val = read_word();
		for (size_t i = 0; i < 16; ++i) {
			bits.push_back( val & (1u<<i) );
		}
		return bits;
	}

	/// Read a bit-array, from 32-bit unsigned.
	std::vector<bool> read_bits32() {
		std::vector<bool> bits;
		bits.reserve(32);
		const unsigned val = read_dword();
		for (size_t i = 0; i < 32; ++i) {
			bits.push_back( val & (1u<<i) );
		}
		return bits;
	}

	/// Read a bit-array, from two 32-bit unsigned ints.
	std::vector<bool> read_bits64() {
		std::vector<bool> bits;
		bits.reserve(64);

		// XXX Note BotF encodes 64-bit bit-arrays as two little-endian
		//     dwords, so we have to read the halves separately.

		// Read low bytes.
		const unsigned lo = read_dword();
		for (size_t i = 0; i < 32; ++i) {
			bits.push_back( lo & (1u<<i) );
		}

		// Read high bytes.
		const unsigned hi = read_dword();
		for (size_t i = 0; i < 32; ++i) {
			bits.push_back( hi & (1u<<i) );
		}

		return bits;
	}


	/// Read a 32-bit floating point number.
	float read_float() {
		union 
		{
			Uint32 ui;
			float f;
		} swapper;
		read( &swapper.ui, 4 );
		swapper.ui = SDL_Swap32( swapper.ui );
		return swapper.f;
	}

	/// Read a 64-bit floating point number.
	double read_double() {
		union 
		{
			Uint64 ui;
			double d;
		} swapper;
		read( &swapper.ui, 8 );
		swapper.ui = SDL_Swap64( swapper.ui );
		return swapper.d;
	}


	/// Reads a number of bytes, returning them as a vector<unsigned char>.
	std::vector<char> read (size_t length) {
		assert (length < 16384); // Don't let bugs write insane data.
		std::vector<char> vec( length, 0 );
		read( &vec[0], length );
		return vec;
	}

	/// Read a fixed-length string (NUL-terminated by default).
	std::string read_string (size_t length, char terminator='\0') 
	{
		assert (length < 16384); // Don't let bugs write insane data.
		// Read exactly 'length' bytes.
		const std::vector<char> vec( read( length ) );
		// Find a NUL-terminator.
		std::vector<char>::const_iterator nul_it = 
			std::find( vec.begin(), vec.end(), terminator );
		// Copy data up to the NUL-terminator into the string.
		std::string str( vec.begin(), nul_it );
		// Return the string.
		return str;
	}

	/// \}

};

#endif

