/* Copyright 2013 Daniel Hanrahan. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE DANIEL HANRAHAN "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL DANIEL HANRAHAN BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef BOF_LIL_ENDIAN_WRITER_H
#define BOF_LIL_ENDIAN_WRITER_H

#include <string>
#include <vector>
#include <algorithm>
#include <boost/smart_ptr/shared_ptr.hpp>
#include <SDL_endian.h>
#ifdef __MINGW32__
#undef main // Prevents SDL from overriding main()
#endif

/// Interface for writing data with little endian encoding.
class BofWriter {
	// Disable copy and assignment.
	BofWriter (const BofWriter& other);
	const BofWriter& operator= (const BofWriter& other);

	/// \ingroup Binary output:
	/// Any subclass MUST implement this function!
	virtual void write (const void *data, size_t n) = 0;

public:
	typedef boost::shared_ptr<BofWriter> ptr;

	BofWriter() {}
	virtual ~BofWriter() {}

	/// How big is the buffer?
//	virtual size_t size() const = 0;


	/// \name Positioning:
	/// \{

	/// How far are we from the start of the buffer?
	virtual size_t position() const = 0;

	/// Set the position in the buffer.
	virtual void seek (size_t position) = 0;

	/// \}


	/// \name Binary output:
	/// \{

	/// Write n nul bytes to the buffer.
	//void write_zeroes (size_t length) {
	void pad (const size_t length) {
		assert (length < 16384); // Don't let bugs write insane data.
		for (size_t i = 0; i < length; ++i) {
			write_byte( 0 );
		}
	}

	/// Write a byte.
	void write_byte (Uint8 val) {
		write( &val, 1 );
	}

	/// Write a character.
	void write_char (Sint8 val) {
		write( &val, 1 );
	}

	/// Write a 16-bit unsigned int.
	void write_word (Uint16 val){
		val = SDL_SwapLE16( val );
		write( &val, 2 );
	}

	/// Write a 16-bit signed int.
	void write_short (Sint16 val) {
		union
		{
			Sint16 si;
			Uint16 ui;
		} swapper;
		swapper.si = val;
		swapper.ui = SDL_SwapLE16( swapper.ui );
		write( &swapper.ui, 2 );
	}

	/// Write a 32-bit unsigned int.
	void write_dword (Uint32 val) {
		val = SDL_SwapLE32( val );
		write( &val, 4 );
	}

	/// Write a 32-bit signed int.
	void write_int (Sint32 val) {
		union
		{
			Sint32 si;
			Uint32 ui;
		} swapper;
		swapper.si = val;
		swapper.ui = SDL_SwapLE32( swapper.ui );
		write( &swapper.ui, 4 );
	}


	/// Write a boolean, 8-bit unsigned.
	void write_bool8 (bool b) {
		Uint8 val = ( b ? 1 : 0 );
		write_byte( val );
	}

	/// Write a boolean, 16-bit unsigned.
	void write_bool16 (bool b) {
		Uint16 val = ( b ? 1 : 0 );
		write_word( val );
	}

	/// Write a boolean, 32-bit unsigned.
	void write_bool32 (bool b) {
		Uint32 val = ( b ? 1 : 0 );
		write_dword( val );
	}


	/// Write a bit-array, as 8-bit unsigned.
	void write_bits8 (const std::vector<bool>& bits) {
		assert( bits.size() <= 8 );
		Uint8 val = 0;
		for (size_t i = 0; i < bits.size(); ++i) {
			if (bits[i]) val |= (1u<<i);
		}
		write_byte( val );
	}

	/// Write a bit-array, as 16-bit unsigned.
	void write_bits16 (const std::vector<bool>& bits) {
		assert( bits.size() <= 16 );
		Uint16 val = 0;
		for (size_t i = 0; i < bits.size(); ++i) {
			if (bits[i]) val |= (1u<<i);
		}
		write_word( val );
	}

	/// Write a bit-array, as 32-bit unsigned.
	void write_bits32 (const std::vector<bool>& bits) {
		assert( bits.size() <= 32 );
		Uint32 val = 0;
		for (size_t i = 0; i < bits.size(); ++i) {
			if (bits[i]) val |= (1u<<i);
		}
		write_dword( val );
	}

	/// Write a bit-array, as two 32-bit unsigned ints.
	void write_bits64 (const std::vector<bool>& bits) {
		assert( bits.size() <= 64 );

		// XXX Note BotF encodes 64-bit bit-arrays as two little-endian
		//     dwords, so we have to read the halves separately.

		Uint32 lo = 0;
		Uint32 hi = 0;

		// Assemble the low part.
		const size_t lo_size = std::min( bits.size(), (size_t)32 );
		for (size_t i = 0; i < lo_size; ++i) {
			if (bits[i]) lo |= (1u<<i);
		}

		// Assemble the high part.
		if (bits.size() > 32) {
			for (size_t i = 32; i < bits.size(); ++i) {
				if (bits[i]) hi |= (1u<<(i-32));
			}
		}

		write_dword( lo );
		write_dword( hi );
	}


	/// Write a 32-bit floating point number.
	void write_float (float val) {
		union 
		{
			float f;
			Uint32 ui;
		} swapper;
		swapper.f = val;
		swapper.ui = SDL_SwapLE32( swapper.ui );
		write( &swapper.ui, 4 );
	}

	/// Write a 64-bit floating point number.
	void write_double (double val) {
		union 
		{
			double d;
			Uint64 ui;
		} swapper;
		swapper.d = val;
		swapper.ui = SDL_SwapLE64( swapper.ui );
		write( &swapper.ui, 8 );
	}


	/// Writes a number of bytes from a vector.
	void write (const std::vector<char>& data) {
		write( &data[0], data.size() );
	}

	/// Write a fixed-length string (NUL-terminated by default).
	void write_string (const std::string& text, size_t length, 
			char terminator='\0') 
	{
		assert (length < 16384); // Don't let bugs write insane data.

		// Warn on text.length() >= length
		assert( text.size() <= length );

		std::vector<char> data( text.begin(), text.end() );

		// Resize the data buffer to the exact length.
		data.resize( length, terminator );

		// Write it.
		write( data );
	}

	/// \}

};

#endif

