/* Copyright 2013 Daniel Hanrahan. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE DANIEL HANRAHAN "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL DANIEL HANRAHAN BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <bitset>
#include "Edifice.hpp"
#include "TechMap.hpp"
#include "FileSystem.hpp"
#include "ZipFile.hpp"
#include "FilePatcher.hpp"

Edifice::Edifice (const std::string& BofLocation) : 
	m_BofLocation( FixPath (BofLocation) ), 
	m_StBofRes( m_BofLocation + "stbof.res" ), 
	m_TrekExe( m_BofLocation + "trek.exe" )
{
	FileMustExist (m_StBofRes);
	FileMustExist (m_TrekExe);

	ZipFile zip (m_StBofRes);
	this->read (*zip.open_read ("edifice.bst"));
}

void Edifice::patch ()
{
	ZipFile zip (m_StBofRes);
	this->write (*zip.open_write ("edifice.bst"));

	FilePatcher p (m_TrekExe);
	patchTrekExe (p);
}

void Edifice::read (BofReader& r) 
{
	const size_t NumBuildings = r.read_dword();
	const size_t oldMapOffset = r.read_dword();
//	assert (oldMapOffset + 4 * NumBuildings == r.size());
	assert (r.size() > 8 + getIgnoreListsSize() + NumBuildings * (0xB0 + 4));

	// Compute tech map size:
	const size_t oldTechMapSize = computeTechMapSize (r.size(), NumBuildings);

	// And compute number of tech levels:
	m_TechLevels = computeTechLevelsFromTechMapSize (oldTechMapSize);

	// Skip the tech map (we'll rebuild it from scratch when writing):
	r.skip (oldTechMapSize);

	// Skip the ignore lists:
	r.skip (getIgnoreListsSize());

	// Read the buildings:
	readBuildings (r, NumBuildings);

	// Skip the building address map:
//	r.skip (4 * NumBuildings);

	// We should be at the end now.
//	assert (r.eof());
}

void Edifice::write (BofWriter& w) 
{
	// Sort the buildings by tech and rebuild the address map:
	sortBuildings ();
	computeBuildingAddresses ();

	w.write_dword (m_Buildings.size());
	w.write_dword (getMapAddress());
	writeTechMap (w);
	writeIgnoreLists (w);
	writeBuildings (w);
	writeBuildingAddresses (w);
}

size_t Edifice::computeTechMapSize (size_t FileSize, size_t NumBuildings) const 
{
	return FileSize
		- 8			// header
		- getIgnoreListsSize()	// ignore lists
		- NumBuildings * 0xB0 	// buildings
		- NumBuildings * 4	// building address map
		;
}

size_t Edifice::computeTechLevelsFromTechMapSize (const size_t TechMapSize) const 
{
	return ((TechMapSize - 16) / 7 - 2) / 6;
}

int Edifice::getMinBuildingTechLevel() const 
{
	int Min = 255;
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		const int tech = m_Buildings[i].techLevel();
		if (tech < Min) Min = tech;
	}
	return Min;
}

int Edifice::getMaxBuildingTechLevel() const 
{
	int Max = 0;
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		const int tech = m_Buildings[i].techLevel();
		if (tech > Max) Max = tech;
	}
	return Max;
}

int Edifice::getMinTechShift () const 
{
	return -getMinBuildingTechLevel();
}

int Edifice::getMaxTechShift () const 
{
	return 64 - getMaxBuildingTechLevel();
}

void Edifice::shiftTechRequirements (int change) 
{
	assert (change >= getMinTechShift());
	assert (change <= getMaxTechShift());
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		const int tech = m_Buildings[i].techLevel();
		m_Buildings[i].techLevel (tech + change);
	}
}

void Edifice::patchTrekExe (FilePatcher& p) 
{
	patchMaxTechLevel (p);
	patchMaxBuildingNumber (p);
}

void Edifice::patchMaxTechLevel (FilePatcher& p) 
{
	p.seek (0x4f52b);
	p.write ("\x69\xC2", 2);
	p.write_dword (6 * m_TechLevels + 2);

	p.seek (0x4f531);
	p.write_byte (0x90);

	p.seek (0x4f538);
	p.write ("\x90\x90\x90", 3);

	p.seek (0x4f541);
	p.write ("\x66\x8B\x84\x42", 4);
	p.write_dword (24 + m_TechLevels * 4);
	p.write ("\x59\x5B\xC3", 3);

	p.seek (0x4f51f);
	p.write_byte (m_TechLevels);

	p.seek (0x4f822);
	p.write_byte (m_TechLevels);

	p.seek (0x4fa5a);
	p.write_dword (getTechMapSize() + 8);

	p.seek (0xd0d1e);
	p.write_byte (m_TechLevels);

	p.seek (0xd2c70);
	p.write_byte (m_TechLevels);

	p.seek (0xd577a);
	p.write_byte (m_TechLevels);

	//XXX Experimental stuff
//	p.seek (0x29a6c);
//	p.write_byte (m_TechLevels - 1);
}

void Edifice::patchMaxBuildingNumber (FilePatcher& p) 
{
	// In the following codes 3D 01 is actually 013Dh = 317.
	// Patch to (NumberOfBuildings - 1), with the bytes reversed.
	size_t locations1[22] = {
		0x4f8e5, 0x4f90e, 0x4fad8, 0x53105, 
		0x851c3, 0x85296, 0x85345, 0x853d0, 
		0x8544f, 0x85475, 0x85926, 0x85f39, 
		0x8602b, 0x86240, 0x862e3, 0x8636e, 
		0x86394, 0x86755, 0x8680f, 0x868ec, 
		0x869cd, 0x86abd
	};
	for (size_t i = 0; i < 22; ++i) {
		p.seek (locations1[i]);
		p.write_word (m_Buildings.size() - 1);
	}

	// In the following codes 3E 01 is actually 013Eh = 318.
	// Patch to NumberOfBuildings, with the bytes reversed.
	size_t locations2[4] = {
		0xaddc, 0xcb55, 0x39e65, 0xf38c9
	};
	for (size_t i = 0; i < 4; ++i) {
		p.seek (locations2[i]);
		p.write_word (m_Buildings.size());
	}
}

void Edifice::writeTechMap (BofWriter& w) 
{
	TechMap techMap (m_TechLevels);

	// Get the building counts for tech fields and levels.
	for (size_t i = 0; i < m_Buildings.size(); ++i) 
	{
		const size_t fld = m_Buildings[i].techField();
		const size_t lvl = m_Buildings[i].techLevel();
		if (lvl < m_TechLevels) {
			techMap.level(fld, lvl).buildingCount++;
		}
	}

	// Get the tech level building start addresses.
	for (size_t fld = 0; fld < 7; ++fld) 
	for (size_t lvl = 0; lvl < m_TechLevels; ++lvl) 
	{
		// Search the building list for the first building
		// with the required tech field/level.
		for (size_t i = 0; i < m_Buildings.size(); ++i) {
			if (m_Buildings[i].techField() == fld
			 && m_Buildings[i].techLevel() == lvl)
			{
				// Store the address of this building.
				const size_t addr = m_BuildingAddresses[i];
				techMap.level(fld, lvl).startAddress = addr;
				break;
			}
		}
	}

	// Write it all out:
	techMap.write (w);
}

void Edifice::writeIgnoreLists (BofWriter& w) 
{
	for (size_t empire = 0; empire < 5; ++empire) 
	{
		// Build the ignore list for this empire:
		std::bitset<384> ignoreList;
		for (size_t i = 0; i < m_Buildings.size(); ++i) {
			const Building& bld = m_Buildings[i];
			if (bld.id() < ignoreList.size()) {
				ignoreList[bld.id()] = bld.excludeMajors(empire);
			}
		}

//		printf ("%lu) %s\n\n", empire, ignoreList.to_string().c_str());

		// Write it out as bytes:
		for (size_t n = 0; n < ignoreList.size(); n+=8) {
			unsigned char byte = 0;
			for (size_t b = 0; b < 8; ++b) {
				if (ignoreList[n + b]) byte |= (1<<b);
			}
			w.write_byte (byte);
		}
	}
}

void Edifice::readBuildings (BofReader& r, size_t numBuildings) 
{
	m_Buildings.reserve (numBuildings);
	for (size_t i = 0; i < numBuildings; ++i) {
		Building bld;
		bld.read (r);
		m_Buildings.push_back (bld);
	}
}

void Edifice::writeBuildings (BofWriter& w) 
{
	assert (w.position() == getBuildingsOffset());
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		m_Buildings[i].write (w);
	}
}

void Edifice::computeBuildingAddresses () 
{
	const size_t offset = getBuildingsOffset();
	m_BuildingAddresses.clear ();
	m_BuildingAddresses.resize (m_Buildings.size());
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		const size_t addr = 0xB0 * i + offset;
		const size_t id = m_Buildings[i].id();
		assert (m_BuildingAddresses.at(id) == 0);
		m_BuildingAddresses.at(id) = addr;
	}
}

void Edifice::writeBuildingAddresses (BofWriter& w) 
{
	assert (m_BuildingAddresses.size() == m_Buildings.size());
	for (size_t i = 0; i < m_BuildingAddresses.size(); ++i) {
		size_t addr = m_BuildingAddresses[i];
		w.write_dword (addr);
	}
}

struct BuildingComparitor {
	bool operator() (const Building& a, const Building& b) { 
		if (a.techField() < b.techField()) {
			return true;
		} else if (a.techField() > b.techField()) {
			return false;
		// If tech fields are equal, check tech level:
		} else if (a.techLevel() < b.techLevel()) {
			return true;
		} else {
			return false;
		}
	}
};

void Edifice::sortBuildings () 
{
	std::sort (m_Buildings.begin(), m_Buildings.end(), 
			BuildingComparitor());
}

Building& Edifice::getBuildingById (int id) 
{
	assert (id >= 0);
	for (size_t i = 0; i < m_Buildings.size(); ++i) {
		if (m_Buildings[i].id() == id) {
			return m_Buildings[i];
		}
	}
	throw std::runtime_error("cannot get structure by id");
}

