/* Copyright 2013 Daniel Hanrahan. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE DANIEL HANRAHAN "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL DANIEL HANRAHAN BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// This is a Zip file reader/writer using 7zip.

#include "ZipFile.hpp"
#include "FileReader.hpp"
#include "FileWriter.hpp"
#include "FileSystem.hpp"

ZipFile::ZipFile (const std::string& zipfile) : 
	m_ZipFile( zipfile )
{
	FileMustExist( m_ZipFile );

	// ensure the system processor works
	if (system (NULL) == 0) {
		throw std::runtime_error
			("System command processor not available");
	}
}


// A subclass of FileReader that stores extra metadata.
class ZipFile::ZipReader : public FileReader {
public:
	ZipReader (const std::string& name) :
		FileReader( name ), 
		m_Name( name )
	{ /*empty*/ }

	std::string name() const { return m_Name; }

private:
	const std::string m_Name;
};

// A custom deleter that removes the temp file.
void ZipFile::ZipReaderDeleter (BofReader *br)  // static
{
	ZipReader *zr = static_cast<ZipReader*>(br);
	const std::string Name = zr->name();
	zr = NULL;

	// Free the Reader object, and ensure file is written.
	delete br;

	// Delete the file:
	remove (Name.c_str());
}

// Open a file for reading.
boost::shared_ptr<BofReader>  ZipFile::open_read (const std::string& name)
{
	// Unzip the file:
	std::string cmd = "7za x -y -bd -tzip " + m_ZipFile + ' ' + name;
	if (system (cmd.c_str()) != 0) {
		throw std::runtime_error("Error unzipping " + name);
	}

	return boost::shared_ptr<BofReader>
		(new ZipReader (name), ZipReaderDeleter);
}


// A subclass of FileWriter that stores extra metadata.
class ZipFile::ZipWriter : public FileWriter {
public:
	ZipWriter (const std::string& zipfile, const std::string& name) :
		FileWriter( name ), 
		m_ZipFile( zipfile ), 
		m_Name( name )
	{ /*empty*/ }

	std::string zipfile() const { return m_ZipFile; }
	std::string name() const { return m_Name; }

private:
	const std::string m_ZipFile;
	const std::string m_Name;
};

// A custom deleter that updates the Zip file and removes the temp file.
void ZipFile::ZipWriterDeleter (BofWriter *bw)  // static
{
	ZipWriter *zw = static_cast<ZipWriter*>(bw);
	const std::string ZipFile = zw->zipfile();
	const std::string Name = zw->name();
	zw = NULL;

	// Free the Writer object, and ensure file is written.
	delete bw;

	// Update the zip file:
	std::string cmd = "7za u -y -bd -tzip " + ZipFile + ' ' + Name;
	if (system (cmd.c_str()) != 0) {
		throw std::runtime_error("Error zipping " + Name);
	}

	// Delete the file:
	remove (Name.c_str());
}

// Open a file for writing.
boost::shared_ptr<BofWriter>  ZipFile::open_write (const std::string& name)
{
	return boost::shared_ptr<BofWriter>
		(new ZipWriter (m_ZipFile, name), ZipWriterDeleter);
}

